use std::{fmt::Display, ops::{Deref, DerefMut}};

use bonsaidb::core::key::{Key, KeyEncoding};
use serde::{de::DeserializeOwned, Deserialize, Deserializer, Serialize};

#[derive(Debug)]
pub(crate) struct KeyEncodingError;

impl Display for KeyEncodingError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Error in handling a bonsaidb-encoded key")
    }
}

impl std::error::Error for KeyEncodingError {}


pub(crate) trait SerdeCompatibleKey: Serialize + DeserializeOwned + Default + Clone + Send + Sync {}
impl<T> SerdeCompatibleKey for T where T: Serialize + DeserializeOwned + Default + Clone + Send + Sync {}

#[derive(Clone, Debug, Default, Serialize, PartialEq, Eq, PartialOrd, Ord)]
pub(crate) struct AsKey<T: SerdeCompatibleKey>(T);

impl<T: SerdeCompatibleKey> AsKey<T>
{
    pub fn new(key: T) -> Self {
        AsKey(key)
    }
}

impl<T: SerdeCompatibleKey> Deref for AsKey<T> {
    type Target = T;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl<T: SerdeCompatibleKey> DerefMut for AsKey<T> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.0
    }
}

impl<'de, T> Deserialize<'de> for AsKey<T>
where
    T: SerdeCompatibleKey,
{
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: Deserializer<'de>,
    {
        let t = T::deserialize(deserializer)?;
        Ok(AsKey(t))
    }
}

impl<'k, T: SerdeCompatibleKey> Key<'k> for AsKey<T>
{
    const CAN_OWN_BYTES: bool = false;

    fn first_value() -> Result<Self, bonsaidb::core::key::NextValueError> {
        Ok(AsKey::default())
    }

    fn from_ord_bytes<'e>(
        bytes: bonsaidb::core::key::ByteSource<'k, 'e>,
    ) -> Result<Self, Self::Error> {
        match serde_json::from_slice(&*bytes) as Result<T, _> {
            Ok(k) => Ok(AsKey(k)),
            Err(_) => Err(KeyEncodingError),
        }
    }
}

impl<T: SerdeCompatibleKey> KeyEncoding for AsKey<T>
{
    type Error = KeyEncodingError;
    const LENGTH: Option<usize> = None;

    fn describe<Visitor>(visitor: &mut Visitor)
    where
        Visitor: bonsaidb::core::key::KeyVisitor,
    {
        visitor.visit_type(bonsaidb::core::key::KeyKind::Bytes)
    }

    fn as_ord_bytes(&self) -> Result<std::borrow::Cow<'_, [u8]>, Self::Error> {
        Ok(std::borrow::Cow::Owned(
            serde_json::to_vec(&self.0).unwrap(),
        ))
    }
}
