use anyhow::{anyhow, Error};
use bonsaidb::{
    core::schema::{Collection, SerializedCollection},
    local::Database,
};
use serde::{Deserialize, Serialize};
use ubisync_lib::types::{
    ContentUpdateStrategy, Element, ElementContent, ElementId, MessageId, PotId,
};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "elements", views = [])]
pub(super) struct DbElement {
    #[natural_id]
    pub(super) id: AsKey<ElementId>,
    pub(super) content: ElementContent,
    pub(super) update_strategy: ContentUpdateStrategy,
    pub(super) latest_message: Option<MessageId>,
    pub(super) local_changes: bool,
    pub(super) pot: PotId,
}

impl From<DbElement> for Element {
    fn from(value: DbElement) -> Self {
        Element {
            id: (*value.id).clone(),
            content: value.content,
            update_strategy: value.update_strategy,
            latest_message: value.latest_message,
            local_changes: value.local_changes,
            pot: Some(value.pot),
        }
    }
}

pub(crate) struct Elements<'a> {
    parent: &'a StateDB,
    db: &'a Database,
}

impl<'a> Elements<'a> {
    pub const fn new(parent: &'a StateDB, bonsai: &'a Database) -> Self {
        Self { parent, db: bonsai }
    }
    pub fn add(
        &self,
        id: ElementId,
        content: ElementContent,
        update_strategy: ContentUpdateStrategy,
        latest_message: Option<MessageId>,
        local_changes: bool,
        pot: PotId,
    ) -> anyhow::Result<()> {
        DbElement::push(
            DbElement {
                id: AsKey::new(id),
                content,
                update_strategy,
                latest_message,
                local_changes,
                pot,
            },
            self.db,
        )
        .map(|_| ())
        .map_err(|e| anyhow!(e))
    }

    pub fn get(&self, id: ElementId) -> Option<Element> {
        DbElement::get(&AsKey::new(id), self.db)
            .ok()?
            .map(|el| el.contents.into())
    }

    pub fn set_content(&self, id: ElementId, content: ElementContent) -> anyhow::Result<()> {
        DbElement::get(&AsKey::new(id), self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find element by id"))?
            .modify(self.db, |t| t.contents.content = content.clone())
            .map_err(|e| anyhow!(e))
    }

    pub fn set_latest_message(
        &self,
        id: ElementId,
        message: Option<MessageId>,
    ) -> anyhow::Result<()> {
        DbElement::get(&AsKey::new(id), self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find element by id"))?
            .modify(self.db, |t| t.contents.latest_message = message.clone())
            .map_err(|e| anyhow!(e))
    }

    pub fn set_local_changes(&self, id: ElementId, local_changes: bool) -> anyhow::Result<()> {
        DbElement::get(&AsKey::new(id), self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find element by id"))?
            .modify(self.db, |t| t.contents.local_changes = local_changes)
            .map_err(|e| anyhow!(e))
    }

    pub fn remove(&self, id: ElementId) -> anyhow::Result<()> {
        DbElement::get(&AsKey::new(id), self.db)
            .map_err(|e| anyhow!(e))?
            .ok_or(Error::msg("Could not find element by id"))?
            .delete(self.db)
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{
        ContentUpdateStrategy, Element, ElementContent, ElementId, MessageId, PotId,
    };

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let pot_id = PotId::new();
        let element_id = ElementId::new();
        db.elements()
            .add(
                element_id.clone(),
                ElementContent::Text("Content!!!".to_string()),
                ContentUpdateStrategy::default(),
                None,
                false,
                pot_id.clone(),
            )
            .unwrap();

        let retrieved_element = db.elements().get(element_id.clone());

        assert_eq!(
            Some(Element {
                id: element_id,
                content: ElementContent::Text("Content!!!".to_string()),
                update_strategy: ContentUpdateStrategy::default(),
                latest_message: None,
                local_changes: false,
                pot: Some(pot_id),
            }),
            retrieved_element
        )
    }

    #[test]
    fn set_content() {
        let db = StateDB::init(None);
        let element_id = ElementId::new();
        db.elements()
            .add(
                element_id.clone(),
                ElementContent::Text("Content!!!".to_string()),
                ContentUpdateStrategy::default(),
                None,
                false,
                PotId::new(),
            )
            .unwrap();
        db.elements()
            .set_content(
                element_id.clone(),
                ElementContent::Text("New Content!!!".to_string()),
            )
            .unwrap();

        assert_eq!(
            db.elements().get(element_id).unwrap().content().to_owned(),
            ElementContent::Text("New Content!!!".to_string())
        )
    }

    #[test]
    fn set_latest_message() {
        let db = StateDB::init(None);

        let element_id = ElementId::new();
        db.elements()
            .add(
                element_id.clone(),
                ElementContent::Text("Content!!!".to_string()),
                ContentUpdateStrategy::default(),
                None,
                false,
                PotId::new(),
            )
            .unwrap();

        assert_eq!(
            db.elements()
                .get(element_id.clone())
                .unwrap()
                .latest_message()
                .to_owned(),
            None
        );

        let msg_id = MessageId::new();
        db.elements()
            .set_latest_message(element_id.clone(), Some(msg_id.clone()))
            .unwrap();

        assert_eq!(
            db.elements()
                .get(element_id)
                .unwrap()
                .latest_message()
                .to_owned(),
            Some(msg_id)
        )
    }

    #[test]
    fn set_local_changes() {
        let db = StateDB::init(None);

        let element_id = ElementId::new();
        db.elements()
            .add(
                element_id.clone(),
                ElementContent::Text("Content!!!".to_string()),
                ContentUpdateStrategy::default(),
                None,
                false,
                PotId::new(),
            )
            .unwrap();

        assert_eq!(
            db.elements()
                .get(element_id.clone())
                .unwrap()
                .local_changes(),
            false
        );

        db.elements()
            .set_local_changes(element_id.clone(), true)
            .unwrap();

        assert_eq!(db.elements().get(element_id).unwrap().local_changes(), true)
    }
}
