use anyhow::{anyhow, Error};
use bonsaidb::{
    core::{
        connection::Connection,
        document::Emit,
        schema::{Collection, MapReduce, SerializedCollection, View, ViewSchema},
    },
    local::Database,
};
use itertools::Itertools;
use serde::{Deserialize, Serialize};
use tracing::debug;
use ubisync_lib::types::{AppId, PotId};

use crate::state::database::{as_key::AsKey, StateDB};

#[derive(Debug, Serialize, Deserialize, Collection, PartialEq, Clone)]
#[collection(name = "pot_memberships", primary_key = u128, views = [DbPotMembershipsByPotId, DbPotMembershipsByAppId, DbPotMembershipsByBothIds])]
pub(super) struct DbPotMembership {
    pub(super) pot_id: AsKey<PotId>,
    pub(super) app_id: AsKey<AppId>,
}

#[derive(Debug, Clone, View, ViewSchema)]
#[view(collection = DbPotMembership, key = AsKey<PotId>, value = Vec<AppId>, name = "by-pot-id")]
pub(super) struct DbPotMembershipsByPotId;

impl MapReduce for DbPotMembershipsByPotId {
    fn map<'doc>(
        &self,
        document: &'doc bonsaidb::core::document::BorrowedDocument<'_>,
    ) -> bonsaidb::core::schema::ViewMapResult<'doc, Self> {
        let entry = DbPotMembership::document_contents(document)?;
        document
            .header
            .emit_key_and_value(entry.pot_id, vec![(*entry.app_id).clone()])
    }

    fn reduce(
        &self,
        mappings: &[bonsaidb::core::schema::MappedValue<
            Self::MappedKey<'_>,
            <Self::View as View>::Value,
        >],
        _rereduce: bool,
    ) -> Result<<Self::View as View>::Value, bonsaidb::core::Error> {
        Ok(mappings
            .iter()
            .map(|mapping| mapping.value.clone())
            .concat())
    }
}

#[derive(Debug, Clone, View, ViewSchema)]
#[view(collection = DbPotMembership, key = AsKey<AppId>, value = Vec<PotId>, name = "by-app-id")]
pub(super) struct DbPotMembershipsByAppId;

impl MapReduce for DbPotMembershipsByAppId {
    fn map<'doc>(
        &self,
        document: &'doc bonsaidb::core::document::BorrowedDocument<'_>,
    ) -> bonsaidb::core::schema::ViewMapResult<'doc, Self> {
        let entry = DbPotMembership::document_contents(document)?;
        document
            .header
            .emit_key_and_value(entry.app_id, vec![(*entry.pot_id).clone()])
    }

    fn reduce(
        &self,
        mappings: &[bonsaidb::core::schema::MappedValue<
            Self::MappedKey<'_>,
            <Self::View as View>::Value,
        >],
        _rereduce: bool,
    ) -> Result<<Self::View as View>::Value, bonsaidb::core::Error> {
        Ok(mappings
            .iter()
            .map(|mapping| mapping.value.clone())
            .concat())
    }
}

#[derive(Debug, Clone, View, ViewSchema)]
#[view(collection = DbPotMembership, key = (AsKey<PotId>, AsKey<AppId>), value = DbPotMembership, name = "by-both")]
pub(super) struct DbPotMembershipsByBothIds;

impl MapReduce for DbPotMembershipsByBothIds {
    fn map<'doc>(
        &self,
        document: &'doc bonsaidb::core::document::BorrowedDocument<'_>,
    ) -> bonsaidb::core::schema::ViewMapResult<'doc, Self> {
        let content = DbPotMembership::document_contents(document)?;
        document
            .header
            .emit_key_and_value((content.pot_id.clone(), content.app_id.clone()), content)
    }
}

pub(crate) struct PotMemberships<'a> {
    parent: &'a StateDB,
    db: &'a Database,
}

impl<'a> PotMemberships<'a> {
    pub const fn new(parent: &'a StateDB, bonsai: &'a Database) -> Self {
        Self { parent, db: bonsai }
    }
    pub fn add(&self, pot: PotId, app: AppId) -> anyhow::Result<()> {
        if let Err(_) = self.parent.pots().get(pot.clone()) {
            Err(Error::msg(
                "A member was meant to be added to a pot which does not exist.",
            ))
        } else if let Err(_) = self.parent.apps().get(app.clone()) {
            Err(Error::msg(
                "A member app which does not exist was meant to be added to a pot",
            ))
        } else {
            if self.get(pot.clone(), app.clone())?.is_some() {
                Err(Error::msg("Pot membership already exists"))
            } else {
                DbPotMembership::push(
                    DbPotMembership {
                        pot_id: AsKey::new(pot),
                        app_id: AsKey::new(app),
                    },
                    self.db,
                )
                .map(|_| ())
                .map_err(|e| anyhow!(e))
            }
        }
    }

    pub fn get(&self, pot: PotId, app: AppId) -> anyhow::Result<Option<()>> {
        self.db
            .view::<DbPotMembershipsByBothIds>()
            .with_key(&(AsKey::new(pot), AsKey::new(app)))
            .query_with_collection_docs()
            .map(|_| Some(()))
            .map_err(|e| anyhow!(e))
    }

    pub fn get_members(&self, pot: PotId) -> anyhow::Result<Vec<AppId>> {
        self.db
            .view::<DbPotMembershipsByPotId>()
            .with_key(&AsKey::new(pot))
            .reduce()
            .map_err(|e| anyhow!(e))
    }

    pub fn get_memberships(&self, app: AppId) -> anyhow::Result<Vec<PotId>> {
        self.db
            .view::<DbPotMembershipsByAppId>()
            .with_key(&AsKey::new(app))
            .reduce()
            .map_err(|e| anyhow!(e))
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::{AppId, PotId};

    use crate::state::database::StateDB;

    #[test]
    fn add_get() {
        let db = StateDB::init(None);
        let pot_id = PotId::new();
        let app_id = AppId::new();
        db.pots()
            .add(pot_id.clone(), "app_type".to_string())
            .unwrap();
        db.apps()
            .add(
                app_id.clone(),
                "name".to_string(),
                "description".to_string(),
                "app_type".to_string(),
            )
            .unwrap();
        db.pot_memberships()
            .add(pot_id.clone(), app_id.clone())
            .unwrap();

        let retrieved_members = db.pot_memberships().get_members(pot_id.clone()).unwrap();
        assert_eq!(vec![app_id.clone()], retrieved_members);

        let retrieved_memberships = db.pot_memberships().get_memberships(app_id).unwrap();
        assert_eq!(vec![pot_id], retrieved_memberships)
    }
}
