use anyhow::anyhow;
use bonsaidb::{
    core::keyvalue::KeyValue,
    local::{
        config::{Builder, StorageConfiguration},Database
    },
};
use tracing::debug;
use ubisync_lib::types::PeerId;
use uuid::Uuid;

use self::collections::{Apps, Elements, Families, Peers, PotMemberships, Pots, UbisyncSchema};

mod as_key;
mod collections;


pub struct StateDB {
    db: Database,
}

impl StateDB {
    pub fn init(path: Option<String>) -> Self {
        let storage_conf = match path {
            Some(p) => StorageConfiguration::new(p),
            None => StorageConfiguration::new(format!("/tmp/{}", Uuid::new_v4())),
            // None => StorageConfiguration::default().memory_only()
        };
        let db = Database::open::<UbisyncSchema>(storage_conf).unwrap();
        StateDB { db }
    }

    pub fn add_family_join_request(&self, peer: PeerId) {
        let _ = self.db.set_key(peer.to_string(), &"").execute();
        debug!(
            "Added join request: {:?}",
            self.db.get_key(peer.to_string()).query()
        )
    }

    pub fn has_family_join_request(&self, peer: PeerId) -> anyhow::Result<bool> {
        Ok(self
            .db
            .get_key(peer.to_string())
            .query()
            .map_err(|e| anyhow!(e))?
            .is_some())
    }

    pub fn remove_family_join_request(&self, peer: PeerId) -> anyhow::Result<()> {
        self.db
            .delete_key(peer.to_string())
            .map(|_| ())
            .map_err(|e| anyhow!(e))
    }

    pub const fn apps(&self) -> Apps {
        Apps::new(&self, &self.db)
    }
    pub const fn elements(&self) -> Elements {
        Elements::new(&self, &self.db)
    }
    pub const fn families(&self) -> Families {
        Families::new(&self, &self.db)
    }
    pub const fn peers(&self) -> Peers {
        Peers::new(&self, &self.db)
    }
    pub const fn pot_memberships(&self) -> PotMemberships {
        PotMemberships::new(&self, &self.db)
    }
    pub const fn pots(&self) -> Pots {
        Pots::new(&self, &self.db)
    }
}

#[cfg(test)]
mod tests {
    use ubisync_lib::types::PeerId;

    use super::StateDB;

    #[test]
    fn family_join_requests() {
        let db = StateDB::init(None);
        let peer = PeerId::default();
        db.add_family_join_request(peer.clone());

        assert!(db.has_family_join_request(peer).unwrap());
    }
}
